// This is -*- C++ -*-
// $Id: DataSet.cc,v 1.21 1998/10/01 19:25:39 trow Exp $

/* 
 * DataSet.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include <algo.h>
#include "Random.h"
#include "DataSet.h"

DataSet::DataSet() : epsilon_(1e-8), type_(0)
{
  clear();
}

DataSet::DataSet(const DataSet& ds)
{
  copy(ds);
}

DataSet&
DataSet::operator=(const DataSet& ds)
{
  copy(ds);
  return *this;
}

const double*
DataSet::sorted_data() const
{
  if (size() == 0) return 0; 
  if (sdata_.size()) return sdata_.data();

  RCArray<double> sd = data_;
  double* s = sd.data();

  if (sortedness() == 0) {
    ::sort(s, s+sd.size(), less<double>());
  } else if (sortedness() == -1) {
    for(size_t i=0; i<size()/2; ++i) {
      double t = s[i];
      s[i] = s[size()-1-i];
      s[size()-1-i] = t;
    }
  }
    
  // Sometimes you just have to fool the compiler...
  ((DataSet*)this)->sdata_ = sd;

  return (const double*)s;
}

double 
DataSet::min() const { return min_; }
double 
DataSet::max() const { return max_; }

double 
DataSet::range() const { return max_ - min_; }
double
DataSet:: sum() const { return sum_; }
double
DataSet:: mean() const { return size() ? sum_/size() : NAN; }
double
DataSet:: var() const { return size() ? sumsq_/size() : NAN; }
double
DataSet:: sdev() const { return size() ? sqrt(sumsq_/size()) : NAN; }
double
DataSet:: sdevs() const { return size()>1 ? sqrt(sumsq_/(size()-1)) : NAN; }

double
DataSet::percentile(double p) const
{
  const double* s = sorted_data();
  double t = p*(size()-1);
  int i = (int)t;
  return s ? (i+1-t)*s[i] + (t-i)*s[i+1] : NAN;
}

double
DataSet:: med() const { return percentile(0.5); }
double
DataSet:: q1() const { return percentile(0.25); }
double
DataSet:: q3() const { return percentile(0.75); }
double
DataSet:: iqr() const { return q3() - q1(); } // interquartile range

double
DataSet::mean_trimmed(size_t l, size_t r) const
{
  if (l+r >= size())
    return NAN;

  const double* s = sorted_data();
  double tot = sum();
  for(size_t i=0; i<l; ++i)
    tot -= s[i];
  for(size_t i=0; i<r; ++i)
    tot -= s[size()-1-i];
  return tot/(size()-l-r);
}

double
DataSet::mean_winsorized(size_t l, size_t r) const
{
  if (l+r >= size())
    return NAN;

  const double* s = sorted_data();
  double tot = sum();
  for(size_t i=0; i<l; ++i)
    tot -= s[i];
  tot += l*s[l];
  for(size_t i=0; i<r; ++i)
    tot -= s[size()-1-i];
  tot += r*s[size()-1-r];
  return tot/size();
}

double
DataSet::moment(unsigned k, double x) const
{
  if (size() == 0) return NAN;

  const double* d = data();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += pow(d[i]-x,k);
  return sum/size();
}

double
DataSet::gmean() const
{
  if (size() == 0 || !is_positive(min())) return NAN;
  const double* d = data();
  double logsum=0;
  for(size_t i=0; i<size(); ++i)
    logsum += log(d[i]);
  return exp(logsum/size());
}

double
DataSet::hmean() const
{
  if (size() == 0) return NAN;
  double invsum=0;
  const double* d = data();
  for(size_t i=0; i<size(); ++i) {
    if (is_zero(d[i])) return NAN;
    invsum += 1/d[i];
  }
  return is_zero(invsum) ? NAN : size()/invsum;
}

double
DataSet::rms() const
{
  // "My work on free software is motivated by an idealistic goal: spreading
  // freedom and cooperation.  I want to encourage free software to spread,
  // replacing proprietary software which forbids cooperation, and thus
  // make our society better." -rms
  if (size() == 0) return NAN;
  const double* d = data();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += d[i]*d[i];
  return sqrt(sum/size());
}

double
DataSet::meandev() const
{
  if (size() == 0) return NAN;
  const double* d = data();
  double sum=0, m=mean();
  for(size_t i=0; i<size(); ++i)
    sum += fabs(d[i]-m);
  return sum/size();
}

double
DataSet::meddev() const
{
  // This is expensive, and it duplicates code.  And it could be optimized.
  // Yipee!
  if (size() == 0) return NAN;
  double m=med();
  const double* dd = data();
  double* d = new double[size()];
  for(size_t i=0; i<size(); ++i)
    d[i] = fabs(dd[i]-m);
  ::sort(d,d+size(),less<double>());
  double md = size()%2 ? d[size()/2] : (d[size()/2-1]+d[size()/2])/2;
  delete [] d;
  return md;
}

double
DataSet::kurtosis() const
{
  if (size() == 0) return NAN;
  const double* d = data();
  double m=mean(), mom2=0, mom4=0;
  for(size_t i=0; i<size(); ++i) {
    double x = d[i]-m;
    x *= x;
    mom2 += x;
    mom4 += x*x;
  }
  return mom4/(mom2*mom2);
}

double
DataSet::skewness() const
{
  if (size() == 0) return NAN;
  const double* d = data();
  double m=mean(), mom2=0, mom3=0;
  for(size_t i=0; i<size(); ++i) {
    double x = d[i]-m;
    mom2 += x*x;
    mom3 += x*x*x;
  }
  return mom3/pow(mom2,1.5);
}

double 
DataSet::excess_kurtosis() const {
  return kurtosis()-3;
}

double 
DataSet::momental_skewness() const {
  return skewness()/2;
}

/*
  This is the test from pg. 93 of Madansky's
  "Prescriptions for Working Statisticians". 
  It checks for independence, with the alternative being
  AR(1) autoregression.  Does it have a name?
*/
double
DataSet::AR1_independence_z() const
{
  size_t N = size();
  double run=0;
  const double* d = data();
  for(size_t i=1; i<N; ++i) {
    double x = d[i] - d[i-1];
    run += x*x;
  }

  double v = run / sumsq_;
  const double exp_mean = 2.0;
  const double exp_var = 4.0 * (N - 1)/(N*N - 1);

  return (v-exp_mean)/sqrt(exp_var);
}

double
DataSet::autocorr(size_t lag) const
{
  size_t N = size();
  assert(lag < N);
  const double* d = data();
  const double m = mean();
  double run = 0;
  for(size_t i=lag; i<N; ++i)
    run += (d[i]-m)*(d[i-lag]-m);
  
  return (run/(N-lag)) / var();
}

double
DataSet::autocorr_z(size_t lag) const
{
  size_t N = size();
  const double ac = autocorr(lag);
  const double exp_mean = -1.0/(N-1);
  const double exp_var  = 1.0/N;
  return (ac - exp_mean) / sqrt(exp_var);
}

double
DataSet::covar(const DataSet& ds) const
{
  if (size() != ds.size()) return NAN;
  const double* d = data();
  const double* dd = ds.data();
  const double m1=mean(), m2=ds.mean();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += (d[i]-m1)*(dd[i]-m2);
  return sum / size();
}

double 
DataSet::corr(const DataSet& ds) const 
{ 
  return covar(ds)/(sdev()*ds.sdev()); 
}

double
DataSet::pooled_mean(const DataSet& ds) const
{
  if (ds.size() == 0 && size() == 0)
    return NAN;
  return (sum() + ds.sum()) / (size() + ds.size());
}

double
DataSet::pooled_var(const DataSet& ds) const
{
  if (ds.size() == 0 && size() == 0)
    return NAN;
  else if (ds.size() == 0)
    return var();
  else if (size() == 0)
    return ds.var();

  const DataSet* a = size() < ds.size() ? &ds : this;
  const DataSet* b = size() < ds.size() ? this : &ds;

  double sumsq = a->sumsq_;
  double sum = a->sum_;
  size_t count = a->size();

  const size_t N = b->size();
  for(size_t i=0; i<N; ++i) {
    double x = b->data(i);
    sum += x;
    sumsq += (x-sum/(count+1))*(x-(sum-x)/count);
    ++count;
  }

  return sumsq / count;
}

double
DataSet::weighted_mean(const DataSet& wt) const
{
  const size_t N = size();
  if (N == 0 || N != wt.size())
    return NAN;

  const double* d = data();
  const double* w = wt.data();
  double sum = 0.0;
  for (size_t i=0; i<N; ++i)
    sum += d[i] * w[i];

  return sum / wt.sum();
}

size_t
DataSet::greater_than(double x) const
{
  if (size() == 0 || max() <= x) return 0;
  if (min() > x) return size();

  const double* d = sorted_data();
  size_t i=0, j=size()-1;

  for(;;) {
    size_t mid=(i+j)/2;
    if (mid == size()-1) --mid;
    if (d[mid] <= x && x < d[mid+1])
      return size()-mid-1;
    else if (d[mid] > x) 
      j = mid;
    else
      i = mid;
  }
}

size_t
DataSet::less_than(double x) const
{
  if (size() == 0 || min() >= x) return 0;
  if (max() < x) return size();

  const double* d = data();
  size_t i=0, j=size()-1;

  for(;;) {
    size_t mid=(i+j)/2;
    if (mid == size()-1) --mid;
    if (d[mid] < x && x <= d[mid+1])
      return mid+1;
    else if (d[mid] >= x)
      j = mid;
    else
      i = mid;
  }
}

void
DataSet::reserve(size_t R)
{
  data_.reserve(R);
}

void
DataSet::add(double x)
{
  update_add(x);
  data_.add(x);
  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
DataSet::add_at(size_t i, double x)
{
  assert(i <= size());
  update_add(x);
  data_.add_at(i,x);
  recalc_sortedness(); // sortedness will be wrong from update_add().
  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
DataSet::remove_at(size_t i)
{
  assert(i < size());
  update_remove(data(i));
  data_.remove_at(i);
  recalc_sortedness();
  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
DataSet::clear()
{
  data_.clear();
  sdata_.clear();
  min_ = max_ = NAN;
  sum_ = sumsq_ = 0;
  sortedness_ = 2; // constant
}

void
DataSet::append(const DataSet& ds)
{
  // clearly, this could be optimized...
  for(size_t i=0; i<ds.size(); ++i)
    add(ds.data(i));
}

void
DataSet::apply(double (*fn)(double))
{
  double* d = data_.data();
  for(size_t i=0; i<size(); ++i)
    d[i] = fn(d[i]);
  recalc();
}

void
DataSet::linear_transform(double a, double b)
{
  double* d = data_.data();

  for(size_t i=0; i<size(); ++i) 
    d[i] = a*d[i]+b;

  
  double m1 = a*min_+b, m2 = a*max_+b;
  min_ = m1<m2 ? m1 : m2;
  max_ = m1<m2 ? m2 : m1;

  sum_ = a*sum_+b*size();
  sumsq_ *= a*a;
  if (a<0 && sortedness_ != 2)
    sortedness_ = -sortedness_;

  if (sdata_.size()) {
    double* s = sdata_.data();
    // linearly transform our sdata buffer
    for(size_t i=0; i<size(); ++i)
      s[i] = a*s[i]+b;

    // if a<0, reverse the sdata to preserve sortedness
    if (a<0) {
      for(size_t i=0; i<size()/2; ++i) {
	double t=s[i];
	s[i]=s[size()-1-i];
	s[size()-1-i]=t;
	}
    }
  }
}

void
DataSet::log_transform()
{
  if (min() <= 0) {
    clear();
    return;
  }
  double* d = data_.data();
  for (size_t i=0; i<size(); ++i)
    d[i] = log(d[i]);


  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = log(s[i]);
  }

  recalc_saving_sorted();
}

void
DataSet::exp_transform()
{
  double* d = data_.data();
  for (size_t i=0; i<size(); ++i)
    d[i] = exp(d[i]);


  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = exp(s[i]);
  }

  recalc_saving_sorted();
}

void
DataSet::logit_transform()
{
  if (min() <= 0 || max() >= 1) {
    clear();
    return;
  }
  double* d = data_.data();
  for(size_t i=0; i<size(); ++i)
    d[i] = log(d[i]/(1-d[i]));

  // the logit transform is order-preserving, so we can zap our
  // sorted data as well
  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = log(s[i]/(1-s[i]));
  }

  recalc_saving_sorted();
}

void
DataSet::reverse()
{
  if (sortedness_ == 2) return;

  double* d = data_.data();

  for(size_t i=0; i<size()/2; ++i) {
    double t = d[i];
    d[i] = d[size()-i-1];
    d[size()-1-i] = t;
  }
  sortedness_ = -sortedness_;
}

void
DataSet::sort(int dir)
{
  // If we are already sorted properly, ignore
  if (sortedness_ == 2 || 
      (sortedness_ == 1 && dir >= 0) ||
      (sortedness_ == -1 && dir < 0))
    return;

  // If we have sorted data already, use it!
  if (dir >= 0 && sdata_.size()) {
    data_ = sdata_;
    sortedness_ = 1;
    return;
  }

  // If we are being asked to "opposite-sort" already-sorted data, do so
  // via reverse()
  if ((dir >= 0 && sortedness_ == -1) || (dir < 0 && sortedness_ == 1)) {
    reverse();
  } else if (dir >= 0) {
    // Otherwise there is no avoiding it, so just sort the data...
    double* d = data_.data();
    ::sort(d, d+size(), less<double>());
    sortedness_ = 1;
  } else {
    double* d = data_.data();
    ::sort(d, d+size(), greater<double>());
    sortedness_ = -1;
  }

  if (sortedness_ > 0)
    sdata_ = data_;
}

void
DataSet::scramble()
{
  if (size() == 0 || sortedness_ == 2) return;

  // If we were sorted before, copy the sorted data so we don't lose it
  double* d = data_.data();

  for(size_t i=0; i<size()-1; ++i) {
    size_t j = rnd.random_integer(i,size()-1);
    if (i != j) {
      double t = d[i];
      d[i] = d[j];
      d[j] = t;
    }
  }

  recalc_sortedness();
}

void
DataSet::permute(const Permutation& p)
{
  assert(size() == p.size());
  
  double* d = data_.data();
  p.permute(d);

  recalc_sortedness();
}

void
DataSet::rank()
{
  pair<double,int>* rankdata = new pair<double,int>[size()];
  double* d = data_.data();

  for(size_t i=0; i<size(); ++i) {
    rankdata[i].first = d[i];
    rankdata[i].second = i;
  }

  ::sort(rankdata, rankdata+size(), less< pair<double,int> >());

  // set ranks, handling ties by the usual convention
  unsigned count=1;
  for(size_t i=0; i<size(); /* increment down below */ ) {
    size_t j=i+1;
    unsigned count2=0;
    while (j < size() && are_equal(rankdata[i].first, rankdata[j].first)) {
      ++j;
      ++count2;
    }
    while (i != j)
      rankdata[i++].first = (2*count + count2)/2.0; // we increment i here
    count += 1+count2;
  }

  for(size_t i=0; i<size(); ++i)
    d[rankdata[i].second] = rankdata[i].first;

  recalc(); // This could be optimized...

  delete [] rankdata;
}

Permutation
DataSet::sorting_permutation(int dir = 1) const
{
  Permutation p;

  if (sortedness_ == 2 ||
      (dir >= 0 && sortedness_ == 1) ||
      (dir < 0 && sortedness_ == -1)) {

    p.set_identity(size());

  } else if ((dir > 0 && sortedness_ == -1) ||
	     (dir < 0 && sortedness_ == 1)) {

    p.set_reverse(size());

  } else {

    pair<double,size_t>* tmp = new pair<double,size_t>[size()];
    size_t* tmp2 = new size_t[size()];
    const double* d = data_.data();
    for (size_t i=0; i<size(); ++i) {
      tmp[i].first = d[i];
      tmp[i].second = i;
    }

    if (dir >= 0) 
      ::sort(tmp, tmp+size(), less< pair<double,size_t> >());
    else
      ::sort(tmp, tmp+size(), greater< pair<double,size_t> >());

    for(size_t i=0; i<size(); ++i)
      tmp2[tmp[i].second] = i;

    p.initialize(size(), tmp2);
    delete [] tmp;
    delete [] tmp2;
  }

  return p;
}

void
DataSet::update_add(double x)
{
    // update statistics /////

  size_t N = size();

  if (N == 0 || x < min_)
    min_ = x;
  if (N == 0 || x > max_)
    max_ = x;
  
  sum_ += x;
  if (N > 0) {
    sumsq_ += (x-sum_/(N+1))*(x-(sum_-x)/N); // cool, huh?

    // Update sortedness flag
    if (sortedness_) {
      const double* d = data_.data_const();
      double prev = d[N-1];
      if (sortedness_ == 2) { // constant
	if (prev < x)
	  sortedness_ = 1;
	else if (prev > x)
	  sortedness_ = -1;
      } else if (sortedness_ == 1) { // ascending
	if (prev > x)
	  sortedness_ = 0;
      } else if (sortedness_ == -1) { // descending
	if (prev < x)
	  sortedness_ = 0;
      }
    }
  }
}

void
DataSet::update_remove(double x)
{
  size_t N = size()-1;
  if (N > 0)
    sumsq_ -= (x-sum_/(N+1))*(x-(sum_-x)/N);
  sum_ -= x;

  if (N == 0) 
    min_ = max_ = NAN;
  else {
    const double* d = data_.data_const();
    size_t NN = size();
    if (min_ == x) {
      bool skip_x = true;
      min_ = d[0];
      if (x == d[0]) {
	min_ = d[1];
	skip_x = false;
      }
      for(size_t i=1; i<NN; ++i) {
	if (d[i] == x && skip_x) {
	  skip_x = false;
	  continue;
	}
	if (min_ > d[i])
	  min_ = d[i];
      }
    }
    if (max_ == x) {
      bool skip_x = true;
      max_ = d[0];
      if (x == d[0]) {
	max_ = d[1];
	skip_x = false;
      }
      for(size_t i=1; i<NN; ++i) {
	if (d[i] == x && skip_x) {
	  skip_x = false;
	  continue;
	}
	if (max_ < d[i])
	  max_ = d[i];
      }
    }
  }
  // Leaves sortedness is fubared state.
}

// There is some code duplication between these and add(), but calling a
// recalc() should be a lot more efficient than just repeatedly add()-ing
// the numbers if you've been fiddling directly with data_.

void
DataSet::recalc_sortedness()
{
  sortedness_ = 2;
  const double* d = data_.data_const();
  size_t N = data_.size();

  for(size_t i=1; i<N && sortedness_; ++i) {
    double x = d[i];
    double prev = d[i-1];
    if (sortedness_ == 2) {
      if (prev < x)
	sortedness_ = 1;
      else if (prev > x)
	sortedness_ = -1;
    } else if (sortedness_ == 1) {
      if (prev > x)
	sortedness_ = 0;
    } else if (sortedness_ == -1) {
      if (prev < x)
	sortedness_ = 0;
    }
  }
}

void
DataSet::recalc()
{
  min_ = max_ = NAN;
  sum_ = sumsq_ = 0;
  sortedness_ = 2;
  if (size() == 0) return;

  const double* d = data_.data_const();
  size_t N = data_.size();

  min_ = max_ = d[0];
  for (size_t i=0; i<N; ++i) {
    double x = d[i];
    if (min_ > x) min_ = x;
    if (max_ < x) max_ = x;
    sum_ += x;
    if (i>0) {
      sumsq_ += (x-sum_/(N+1))*(x-(sum()-x)/N);
      if (sortedness_) {
	double prev = d[i-1];
	if (sortedness_ == 2) {
	  if (prev < x)
	    sortedness_ = 1;
	  else if (prev > x)
	    sortedness_ = -1;
	} else if (sortedness_ == 1) {
	  if (prev > x)
	    sortedness_ = 0;
	} else if (sortedness_ == -1) {
	  if (prev < x)
	    sortedness_ = 0;
	}
      }
    }
  }

  sdata_.clear();
}

void
DataSet::recalc_saving_sorted()
{
  // A hack to let us call recalc() w/o losing the contents of sdata_.
  RCArray<double> foo = sdata_;
  recalc();
  sdata_ = foo;
}

void
DataSet::copy(const DataSet& ds)
{
  clear();
  label_ = ds.label_;
  data_ = ds.data_;
  sdata_ = ds.sdata_;
  min_ = ds.min_;
  max_ = ds.max_;
  sum_ = ds.sum_;
  sumsq_ = ds.sumsq_;
  sortedness_ = ds.sortedness_;
  epsilon_ = ds.epsilon_;
}



// $Id: DataSet.cc,v 1.21 1998/10/01 19:25:39 trow Exp $

