// This is -*- C++ -*-
// $Id: StatsDB.cc,v 1.1 1998/09/29 23:16:53 hp Exp $

/* 
 * StatsDB.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Havoc Pennington <hp@emccta.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "StatsDB.h"

// We do this static array stuff for easy maintenance. You just add
// the info here when you modify DataSet

struct SingleEntryInfo { 
  StatsDB::single_ds_stat_member f;
  const char* name;
};

struct DoubleEntryInfo {
  StatsDB::double_ds_stat_member f;
  const char* name;
};

static const struct SingleEntryInfo global_single_entries[] = {
  { &DataSet::min,     "Minimum" },
  { &DataSet::max,     "Maximum" },
  { &DataSet::range,   "Range" },
  { &DataSet::sum,     "Sum" },
  { &DataSet::mean,    "Mean" },
  { &DataSet::var,     "Variance" },
  { &DataSet::sdev,    "Standard Deviation" },
  { &DataSet::sdevs,   "Sample Standard Deviation" },
  { &DataSet::med,     "Median" },
  { &DataSet::q1,      "Q1" },
  { &DataSet::q3,      "Q3" },
  { &DataSet::iqr,     "Interquartile Range" },
  { &DataSet::gmean,   "Geometric Mean" },
  { &DataSet::hmean,   "Harmonic Mean" },
  { &DataSet::rms,     "Root Mean Square" },
  { &DataSet::meandev, "Mean Deviation" },
  { &DataSet::meddev,  "Median Deviation" },
  { &DataSet::kurtosis, "Kurtosis" },
  { &DataSet::skewness, "Skewness" },
  { &DataSet::excess_kurtosis, "Excess Kurtosis" },
  { &DataSet::momental_skewness, "Momental Skewness" }
};

static size_t num_single_entries = sizeof(global_single_entries)/sizeof(global_single_entries[0]);

static const struct DoubleEntryInfo global_double_entries[] = {
  { &DataSet::covar,           "Covariance" },
  { &DataSet::corr,            "Correlation" },
  { &DataSet::pooled_mean,     "Pooled Mean" },
  { &DataSet::pooled_var,      "Pooled Variance" },
  { &DataSet::weighted_mean,   "Weighted Mean" }
};

static size_t num_double_entries = sizeof(global_double_entries)/sizeof(global_double_entries[0]);

StatsDB::StatsDB()
{
  size_t i = 0;
  while (i < num_single_entries) {
    add(global_single_entries[i].name, global_single_entries[i].f);
    ++i;
  }

  i = 0;
  while (i < num_double_entries) {
    add(global_double_entries[i].name, global_double_entries[i].f);
    ++i;
  }
}

StatsDB::~StatsDB()
{
  map<string,SingleEntry*>::iterator i = single_entries_.begin();
  while (i != single_entries_.end()) {
    delete (*i).second;
    ++i;
  }

  map<string,DoubleEntry*>::iterator j = double_entries_.begin();
  while (j != double_entries_.end()) {
    delete (*j).second;
    ++j;
  }

  vector<char*>::iterator k = single_keys_.begin();
  while (k != single_keys_.end()) {
    free(*k);
    ++k;
  } 

  k = double_keys_.begin();
  while (k != double_keys_.end()) {
    free(*k);
    ++k;
  } 
}

double StatsDB::statistic(const char* key, const DataSet& ds) const
{
  map<string,SingleEntry*>::const_iterator i = single_entries_.find(key);
  
  if (i == single_entries_.end()) {
    cerr << "Goose: Key " << key << " not found." << endl;
    return NAN;
  }
  else {
    return i->second->apply(ds);
  }
}

double StatsDB::statistic(const char* key, 
			  const DataSet & ds1, const DataSet& ds2) const
{
  map<string,DoubleEntry*>::const_iterator i = double_entries_.find(key);
  
  if (i == double_entries_.end()) {
    cerr << "Goose: Key " << key << " not found." << endl;
    return NAN;
  }
  else {
    return i->second->apply(ds1,ds2);
  }
} 

StatsDB::const_iterator StatsDB::single_begin() const
{
  return single_keys_.begin();
}

StatsDB::const_iterator StatsDB::double_begin() const
{
  return double_keys_.begin();
}


StatsDB::const_iterator StatsDB::single_end() const
{
  return single_keys_.end();
}

StatsDB::const_iterator StatsDB::double_end() const
{
  return double_keys_.end();
}


void StatsDB::add(const char* name, single_ds_stat_func f)
{
  SingleEntry* se = new FunctionSingleEntry(f);
  single_keys_.push_back(strdup(name));
  single_entries_[name] = se;
}

void StatsDB::add(const char* name, double_ds_stat_func f)
{
  DoubleEntry* se = new FunctionDoubleEntry(f);
  double_keys_.push_back(strdup(name));
  double_entries_[name] = se;
}


void StatsDB::add(const char* name, single_ds_stat_member f)
{
  SingleEntry* se = new MemberSingleEntry(f);
  single_keys_.push_back(strdup(name));
  single_entries_[name] = se;
}

void StatsDB::add(const char* name, double_ds_stat_member f)
{
  DoubleEntry* se = new MemberDoubleEntry(f);
  double_keys_.push_back(strdup(name));
  double_entries_[name] = se;
}


// $Id: StatsDB.cc,v 1.1 1998/09/29 23:16:53 hp Exp $
