// This is -*- C++ -*-
// $Id: statfn.cc,v 1.9 1998/10/07 05:47:01 trow Exp $


/* statfn.cc
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * Parts of this code are (C) 1998 Ross Ihaka, and is derived from his
 * Mathlib code that is distributed with R (http://fangorn.ci.tuwien.ac.at/R).
 * These R functions were originally GPLed, but are made available to Goose
 * under the LGPL to Goose via a special exception. (Thanks!)
 * Functions that originated in R are labelled below.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <iostream.h>
#include <math.h>
#include <nan.h>
#include "statfn.h"

double
factorial(int n)
{
  static double lookup[24] = {
    1.,
    1.,
    2.,
    6.,
    24.,
    120.,
    720.,
    5040., 
    40320.,
    362880.,
    3628800.,
    39916800.,
    479001600.,
    6227020800.,
    87178291200.,
    1307674368000.,
    20922789888000.,
    355687428096000.,
    6402373705728000.,
    121645100408832000.,
    2432902008176640000.,
    51090942171709440000.,
    1124000727777607680000.,
    25852016738884976640000.
  };
  if (n < 0)
    return NAN;
  else if (n < 24)
    return lookup[n];
  else
    return exp(log_factorial(n));
}

double
log_factorial(int n)
{
  return n >= 0 ? log_gamma(1.0+n) : NAN;
}

double
choose(int n, int k)
{
  if (n < 1 || k < 0 || k > n)
    return 0;
  else if (k == 0 || k == n)
    return 1;
  else if (k == 1 || k == n-1)
    return n;
  else if (n < 13)
    return factorial(n)/(factorial(k)*factorial(n-k));
  else
    return exp(log_choose(n,k));
}

double
log_choose(int n, int k)
{
  return log_factorial(n) - log_factorial(k) - log_factorial(n-k);
}

double
log_gamma(double x)
{
  double sum = 1.000000000190015
    + 76.18009172947146 / (x+1)
    - 86.50532032941677 / (x+2)
    + 24.01409824083091 / (x+3)
    - 1.231739572450155 / (x+4)
    + 0.1208650973866179e-2 / (x+5)
    - 0.5395239384953e-5 / (x+6);

  return log(2.5066282746310005 * sum / x) - x - 5.5 + (x + 0.5)*log(x + 5.5);
}

// This function is adapted from one in R (see license header above)
double
incomplete_gamma(double p, double x)
{
  const double plimit = 1000.0;
  const double elimit = -88.0;
  const double xbig = 1.0e+8;
  const double EPSILON = 1e-16;
  const double oflo = 1.0e+37;

  double pn1, pn2, pn3, pn4, pn5, pn6, arg, c, rn, a, b, an;
  double sum;

  /* check that we have valid values for x and p */
  
  if (isnan(x) || isnan(p) || p <= 0.0)
    return NAN;

  if (x <= 0.0)
    return 0.0;

  /* use a normal approximation if p > plimit */

  if (p > plimit) {
    pn1 = sqrt(p) * 3.0 * (pow(x/p, 1/3.0) + 1.0 / (p * 9.0) - 1.0);
    return normal_cdf(pn1);
  }

  /* if x is extremely large compared to p then return 1 */

  if (x > xbig)
    return 1.0;

  if (x <= 1.0 || x < p) {

    /* use pearson's series expansion. */
    
    arg = p * log(x) - x - log_gamma(p + 1.0);
    c = 1.0;
    sum = 1.0;
    a = p;
    do {
      a = a + 1.0;
      c = c * x / a;
      sum = sum + c;
    } while (c > EPSILON);
    arg = arg + log(sum);
    sum = 0.0;
    if (arg >= elimit)
      sum = exp(arg);
  } else {
    
    /* use a continued fraction expansion */

    arg = p * log(x) - x - log_gamma(p);
    a = 1.0 - p;
    b = a + x + 1.0;
    c = 0.0;
    pn1 = 1.0;
    pn2 = x;
    pn3 = x + 1.0;
    pn4 = x * b;
    sum = pn3 / pn4;
    for (;;) {
      a = a + 1.0;
      b = b + 2.0;
      c = c + 1.0;
      an = a * c;
      pn5 = b * pn3 - an * pn1;
      pn6 = b * pn4 - an * pn2;
      if (fabs(pn6) > 0.0) {
	rn = pn5 / pn6;
	if (fabs(sum - rn) <= (rn<1 ? rn*EPSILON : EPSILON))
	  break;
		
	sum = rn;
      }
      pn1 = pn3;
      pn2 = pn4;
      pn3 = pn5;
      pn4 = pn6;
      if (fabs(pn5) >= oflo) {
	
	/* re-scale the terms in continued fraction */
	/* if they are large */

	pn1 = pn1 / oflo;
	pn2 = pn2 / oflo;
	pn3 = pn3 / oflo;
	pn4 = pn4 / oflo;
      }
    }
    arg = arg + log(sum);
    sum = 1.0;
    if (arg >= elimit)
      sum = 1.0 - exp(arg);
  }
  return sum;
}

// This function is adapted from one in R (see license header above)
double
incomplete_beta(double pin, double qin, double x)
{
  if (isnan(x) || isnan(pin) || isnan(qin) || pin <= 0.0 || qin <= 0.0)
    return NAN;
  
  if (x <= 0)
    return 0;
  if (x >= 1)
    return 1;

  double ans, c, finsum, p, ps, p1, q, term, xb, xi, y;
  int n, i, ib;

  const double eps = 1e-16;
  static double alneps = log(eps);
  static double sml = DBL_MIN;
  static double alnsml = log(sml);

  y = x;
  p = pin;
  q = qin;

  /* swap tails if x is greater than the mean */

  if (p / (p + q) < x) {
    y = 1 - y;
    p = qin;
    q = pin;
  }

  if ((p + q) * y / (p + 1) < eps) {

    /* tail approximation */

    ans = 0;
    xb = p * log(y<sml ? sml : y) - log(p) - log_beta_function(p, q);
    if (xb > alnsml && y != 0)
      ans = exp(xb);
    if (y != x || p != pin)
      ans = 1 - ans;
  }
  else {

    /* evaluate the infinite sum first.  term will equal */
    /* y^p / beta(ps, p) * (1 - ps)-sub-i * y^i / fac(i) */

    ps = q - floor(q);
    if (ps == 0)
      ps = 1;
    xb = p * log(y) - log_beta_function(ps, p) - log(p);
    ans = 0;
    if (xb >= alnsml) {
      ans = exp(xb);
      term = ans * p;
      if (ps != 1) {
	n = (int)(alneps/log(y));
	if (n < 4) n = 4;
	for(i=1 ; i<= n ; i++) {
	  xi = i;
	  term = term * (xi - ps) * y / xi;
	  ans = ans + term / (p + xi);
	}
      }
    }

    /* now evaluate the finite sum, maybe. */
    
    if (q > 1) {
      xb = p * log(y) + q * log(1 - y) - log_beta_function(p, q) - log(q);
      ib = (int)(xb/alnsml);
      if (ib < 0) ib = 0;
      term = exp(xb - ib * alnsml);
      c = 1 / (1 - y);
      p1 = q * c / (p + q - 1);

      finsum = 0;
      n = (int)q;
      if (q == n)
	n = n - 1;
      for(i=1 ; i<=n ; i++) {
	if (p1 <= 1 && term / eps <= finsum)
	  break;
	xi = i;
	term = (q - xi + 1) * c * term / (p + q - xi);
	if (term > 1) {
	  ib = ib - 1;
	  term = term * sml;
	}
	if (ib == 0)
	  finsum = finsum + term;
      }
      ans = ans + finsum;
    }
    if (y != x || p != pin)
      ans = 1 - ans;
    if (ans > 1.0) ans = 1.0;
    if (ans < 0.0) ans = 0.0;
  }
  return ans;
}

//////////////////////////////////////////////////////////////////////////

double
log_beta_function(double z, double w)
{
  return log_gamma(z)+log_gamma(w)-log_gamma(z+w);
}

double
beta_function(double z, double w)
{
  return exp(log_gamma(z)+log_gamma(w)-log_gamma(z+w));
}


/////////////////////////////////////////////////////////////////////////////

/*
  This implementation is inspired by Majumder & Bhattacharjee's
  algorithm in "Applied Statistics", though it is very different.
*/

double
inv_beta_cdf(double p, double a, double b)
{
  // Validate input parameters.
  if (p < 0 || p >1 || a <= 0 || b <= 0)
    return NAN;

  //if (p < 0.5)
  //return 1-inv_beta_cdf(1-p,b,a);

  // First, we find x, an initial estimate to the correct value.
  // We use the estimate from Abramowitz & Stegun, 26.5.22
  double x;
  if (p < 1e-6)
    x = 0;
  else if (1-p < 1e-6)
    x = 1;
  else {
    double yp = inv_normal_cdf(1-p);
    double h  = 2/(1/(2*a-1) + 1/(2*b-1));
    double lambda = (yp*yp-3)/6;
    double w = yp*sqrt(h+lambda)/h - (1/(2*b-1) - 1/(2*a-1))*(lambda+5/6.0 - (2/3.0)*h);
    x = a/(a+b*exp(2*w));
  }

  // Then we use Newton's method to sharpen up the estimate.
  double beta = beta_function(a,b);
  const int max_iter = 100;
  const double eps = 1e-14;
  
  for(int i=0; i<max_iter; ++i) {
    double fx = incomplete_beta(a,b,x)-p;
    if (fabs(fx) < eps)
      break;
    double dfx = exp((a-1)*log(x)+(b-1)*log(1-x) - log(beta));
    x -= fx/dfx;
  }

  return x;
}

/////////////////////////////////////////////////////////////////////////////

double
normal_cdf(double x)
{
  x /= M_SQRT2;
  return 0.5 * (x >= 0 ? 1 + erf(x) : erfc(-x));
}

double
inv_normal_cdf(double p)
{
  double x = p;

  int sign;
  if (x < 0.5) {
    sign = -1;
  } else {
    sign = 1;
    x = 1-x;
  }

  if (x < 1e-20) return sign*10;

  x = sqrt(-2*log(x));    
  
  const double p0 = -0.322232431088;
  const double p1 = -1.0;
  const double p2 = -0.342242088547;
  const double p3 = -0.0204231210245;
  const double p4 = -0.453642210148e-4;

  const double q0 = 0.0993484626060;
  const double q1 = 0.588581570495;
  const double q2 = 0.531103462366;
  const double q3 = 0.103537752850;
  const double q4 = 0.38560700634e-2;

  double num = (((x*p4+p3)*x+p2)*x+p1)*x+p0;
  double den = (((x*q4+q3)*x+q2)*x+q1)*x+q0;

  // this is our initial guess, which is accurate to about 7 decimal places...
  double x0 = sign*(x+num/den); 

  // ...but the following sharpens our estimate up quite a bit.
  double Z = sqrt(2*M_PI)*(p - normal_cdf(x0))/exp(-x0*x0/2);
  return x0 + (((2*x0*x0 + 1)*Z/3 + x0)*Z/2 + 1)*Z;
}

double
poisson_cdf(double k, double x)
{
  return x <= 0 || k < 1 ? NAN : 1 - incomplete_gamma(k,x);
}

double
chi_square_cdf(double chi2, double df)
{
  return chi2 <= 0 ? 0 : incomplete_gamma(df/2,chi2/2);
}

// A stupid, inefficient version.
double
inv_chi_square_cdf(double p, double df)
{
  const double EPSILON = 1e-12;
  double ch = 2*df;

  while (chi_square_cdf(ch, df) <= p)
    ch += df;

  double a = 0, b = ch, guess = ch/2, s;
  while (fabs(p - (s = chi_square_cdf(guess, df))) > EPSILON) {
    if (p < s)
      b = guess;
    else
      a = guess;
    guess = (a+b)/2;
  }

  return guess;
}

double
students_cdf(double t, double df)
{
  double b = 1 - incomplete_beta(df/2, 0.5, df/(df+t*t));
  return 0.5 + (t >= 0 ? b/2 : -b/2);
}

// A stupid, inefficient version.
double
inv_students_cdf(double p, double df)
{
  const double EPSILON = 1e-9;

  // First, calc normal case
  double z = inv_normal_cdf(p);

  // Now search around that value
  double a = z-2, b = z+2, guess = z, s;

  while (students_cdf(a, df) >= p)
    a -= 2;

  while (students_cdf(b, df) <= p)
    b += 2;

  while (fabs(p - (s = students_cdf(guess, df))) > EPSILON) {
    //cout << a << " " << b << " " << s << " " << p << " " << p-s << endl;
    if (p < s)
      b = guess;
    else
      a = guess;
    guess = (a+b)/2;
  }

  return guess;
}

double
F_cdf(double F, double df1, double df2)
{
  return incomplete_beta(df2/2, df1/2, df2/(df2+df1*F));
}




// $Id: statfn.cc,v 1.9 1998/10/07 05:47:01 trow Exp $
